/********************************************************************************/
/*    H   H Y   Y TTTTT EEEEE  CCC       HYTEC ELECTRONICS LTD                  */
/*    H   H  Y Y    T   E     C          5 Cradock Road,                        */
/*    HHHHH   Y     T   EEE   C          Reading, Berks.    Tel: 0118 9757770	*/
/*    H   H   Y     T   E     C          RG2 0JT            Fax: 0118 9757566	*/
/*    H   H   Y     T   EEEEE  CCC       Web: www.hytec-electronics.co.uk       */
/********************************************************************************/
/********************************************************************************/
/*   __________________________________________________________________________ */
/*   H Y T E C   8 5 2 2  1 6 C h a n n e l  S c a l a r  A s y n  D r i v e r  */
/*           -----------------------------------------------------------        */
/*                                                                              */
/*  Source file name      :-      drvHy8522ip.c                                 */
/*                                                                              */
/*  Initial creation date :-      24-Jan-2012                                   */
/*                                                                              */
/*  Original Developers   :-      Jim Chen						                */
/*                                                                              */
/********************************************************************************/
/*                                                                              */
/*  Description :- This code is for Hytec 8522 16 channel histogram scalar		*/
/*	EPICS support. It is based the latest model 3 asyn port driver code. It is 	*/
/*	osi complaint.		                    									*/
/*                                                                              */
/*  This latest development distinguishes from normal scalar by providing		*/
/*	histogram support for each scalar channel. Also it stores multiple shot 	*/
/*	(trigger) data in the memory.												*/
/*                                                                              */
/*  (C)2012 Hytec Electronics Ltd.                                              */
/*                                                                              */
/********************************************************************************/
/*                                                                              */
/* Modification record :-                                                       */
/*                                                                              */
/*  Date modified |  Person   | Version | Description of modification           */
/* ---------------|-----------|---------|-------------------------------------- */
/*    24/01/12    |  JC		  | 1.0     | Initial creation with new asyn model  */    
/*    13/03/12    |  JC		  | 2.0     | Added support for scaler records		*/    
/*    04/07/12    |  JC		  | 2.1     | Added edm example and fixed mode		*/    
/*                |           |         | switching waveform clear              */    
/*    16/07/12    |  JC		  | 2.2     | Add the first version of preset		*/    
/*                |           |         | scaler support			            */    
/*    25/07/12    |  JC		  | 3.0     | Add full support for preset scaler	*/    
/*                |           |         | that meets the scaler record v3.20    */    
/*                |           |         | requirements. Also in histogram mode, */    
/*                |           |         | number of bins is the same as number  */    
/*                |           |         | of gates setting.                     */    
/*    04/04/13    |  JC		  | 3.1     | Two changes for preset scaler mode	*/
/*    		      |  		  |     	| 1.Added channel 0 counter update when	*/
/*                |           |         | changing from oneshot to autocount. 	*/    
/*                |           |         | 2.Fixed a bug when autocount is on  	*/    
/*                |           |         | going then start the normal count.  	*/    
/*    11/07/13    |  JC		  | 4.0     | Major change to the interface of 		*/
/*                |           |         | initialisation routines. Now each   	*/    
/*                |           |         | mode has its own init function so   	*/    
/*                |           |         | it only passes the params needed.		*/    
/*    24/04/14    |  JC		  | 4.1     | Fixed bin data fetch calculation		*/
/*                |           |         | error   								*/
/*    01/05/14    |  JC		  | 4.2     | Changed the default cycle repeat to	*/
/*                |           |         | single mode							*/
/*    06/05/14    |  JC		  | 4.3     | Added incremental bins return for 	*/
/*                |           |         | histogram mode						*/
/*    28/05/14    |  JC		  | 4.4     | Fixed a bug for waveform records  	*/
/*                |           |         | population							*/
/*    16/09/14    |  JC		  | 4.5     | Sorted out the trigger statue			*/
/*                |Mark Rivers|         | Modified to hold driver table to suit */
/*                |           |         | 64bit Linux						   	*/
/*                |           |         | Fixed counting 0s after reset		   	*/
/*    26/09/14    |  JC		  | 4.6     | Fixed the trigger statue on restart.	*/
/*                |           |         | Mimic the hardware trigger status		*/
/*                |           |         | (not exact accurate!)					*/
/*                |           |         |									   	*/
/********************************************************************************/

#define DRIVER_VERSION	406

/******************/
/* Hytec Module includes */
/******************/
#include "drvHy8522Asyn.h"

#include "scalerRecord.h"
#include "devScalerAsyn.h"
#include "dbEvent.h"    /* for db_post_events() */
#include "dbAccess.h"	/* for DBE_VALUE */

static Hy8522Asyn *driverTable[MAX_CARDS];
static int numCards;

/********************************************************************/
/*														           */
/* Public Sector												  */
/*													             */
/****************************************************************/
/* Constructor:
 *
 * Where: -
 *     portName = asyn port name
 *   carrierNum = carrier index number when adding a carrier card by ipacAddHy8002 call
 *    ipSlotNum = IP slot number on the carrier board (0 = A, 1 = B, 2 = C, 3 = D, 4 = E, 5 = F),
 *    vectorNum = interrupt vector (0 = find one for me),
 *	 	   freq = For internal clock. 0: 25MHz, 1:50MHz, 2:100MHz, 3:200MHz
 *	 sourceMask = bit mask for channel clock source. 0: external. 1: internal defined by freq param
 *	 hwtrigger  = 1 = enable hardware trigger, 0 = only software trigger
 *    hwreset   = 1 = allow hardware reset, 0 = only software reset
*/
    
Hy8522Asyn::Hy8522Asyn(char *portName, int carrierNum, int ipSlotNum, int vectorNum, int freq, int sourceMask, int hwtrigger, int hwreset)    
		: asynPortDriver(portName, MAX_CHANNELS, NUM_PARAMS, 
		asynInt32Mask | asynInt32ArrayMask | asynFloat64ArrayMask | asynDrvUserMask,
        asynInt32Mask | asynInt32ArrayMask | asynFloat64ArrayMask, 
        ASYN_MULTIDEVICE, 1, /* ASYN_CANBLOCK=0, ASYN_MULTIDEVICE=1, autoConnect=1 */
        0, 0)  /* Default priority and stack size */

{
    const char *hytecstr = "(Hytec Electronics Ltd., Reading, UK)";
    int strok;
  	char lstr[7];
    int ishytec;
    int ismodel;
	int ver, i;
	USHORT idmem;
	
	/* normal data channels definition */
    createParam(hy8522DataString, 			asynParamInt32, 	&hy8522Data_);
    createParam(hy8522CoincidenceString, 	asynParamInt32, 	&hy8522Coin_);
    createParam(hy8522ModeString, 			asynParamInt32, 	&hy8522Mode_);
    createParam(hy8522SoftTriggerString, 	asynParamInt32, 	&hy8522SoftTrigger_);

	/* controls definition */
    createParam(hy8522HWTriggerString, 		asynParamInt32, 	&hy8522HWTrigger_);
    createParam(hy8522HWResetString, 		asynParamInt32, 	&hy8522HWReset_);
    createParam(hy8522GateIntervalString, 	asynParamInt32, 	&hy8522GateIntv_);
    createParam(hy8522NoCyclesString, 		asynParamInt32, 	&hy8522NoCycles_);
    createParam(hy8522NoGatesString, 		asynParamInt32, 	&hy8522NoGates_);
    createParam(hy8522DisableInputString, 	asynParamInt32, 	&hy8522DiableInput_);
    createParam(hy8522TimeUnitString, 		asynParamInt32, 	&hy8522TimeUnit_);
    createParam(hy8522ExtAdvanceString,		asynParamInt32, 	&hy8522ExtAdvance_);
    createParam(hy8522AddupString,			asynParamInt32, 	&hy8522Addup_);
    createParam(hy8522FrequencyString, 		asynParamInt32, 	&hy8522Frequency_);
    createParam(hy8522SourceString, 		asynParamInt32, 	&hy8522Source_);
    createParam(hy8522RestartString, 		asynParamInt32, 	&hy8522Restart_);

    createParam(SCALER_RESET_COMMAND_STRING,     	asynParamInt32, &scalerReset_);               /* int32, write */
    createParam(SCALER_CHANNELS_COMMAND_STRING,  	asynParamInt32, &scalerChannels_);            /* int32, read */
    createParam(SCALER_READ_COMMAND_STRING,      	asynParamInt32Array, &scalerRead_);           /* int32Array, read */
    createParam(SCALER_READ_SINGLE_COMMAND_STRING,  asynParamInt32, &scalerReadSingle_);          /* int32, read */
    createParam(SCALER_PRESET_COMMAND_STRING,    	asynParamInt32, &scalerPresets_);     		  /* int32, write */
    createParam(SCALER_ARM_COMMAND_STRING,       	asynParamInt32, &scalerArm_);                 /* int32, write */
    createParam(SCALER_DONE_COMMAND_STRING,      	asynParamInt32, &scalerDone_);                /* int32, read */

    createParam(hy8522MemPointerString,		asynParamInt32, 	&hy8522MemPointer_);
    createParam(hy8522NoTriggersString,		asynParamInt32, 	&hy8522NoTriggers_);
    createParam(hy8522TriggerStatusString,	asynParamInt32, 	&hy8522TriggerStatus_);
    createParam(hy8522FWVersionString, 		asynParamInt32, 	&hy8522FWVersion_);
    createParam(hy8522DriverVersionString,	asynParamInt32, 	&hy8522DriverVersion_);
    createParam(hy8522SupportString, 		asynParamInt32, 	&hy8522Support_);
	
	/* Save the parameters in the driver structure    */
    this->carrierNum_		 = carrierNum;
    this->ipSlotNum_		 = ipSlotNum;
    this->vectorNum_         = vectorNum;
    this->isdone_			 = false;
    
	setIntegerParam(hy8522Mode_, 0);																//initial mode to normal histogram.
	setIntegerParam(hy8522HWTrigger_, hwtrigger);
	setIntegerParam(hy8522HWReset_, hwreset);
	setIntegerParam(hy8522Restart_, hy8522singleshot);
	setIntegerParam(hy8522Frequency_, freq);
	setIntegerParam(hy8522Source_, sourceMask);
	setIntegerParam(hy8522TriggerStatus_, WAITING_TRIGGER);

	// Default values of some parameters
	setIntegerParam(scalerDone_, 0);																//initial clear the done bit.
	setIntegerParam(scalerChannels_, MAX_CHANNELS);													//maxSignals
	for (i=0; i<MAX_CHANNELS; i++) {
	    setIntegerParam(i, scalerPresets_, 0);
	    callParamCallbacks(i);
	}

	setIntegerParam(hy8522ExtAdvance_, 0);															//internal time advancing
	setIntegerParam(hy8522Addup_, 0);																//initial mode to normal histogram.
	setIntegerParam(hy8522Frequency_, 0);															//25MHz by default

	  /* Map Register address spaces  */
	this->scalar = (IO_MEMORY *) ipmBaseAddr(carrierNum, ipSlotNum, ipac_addrIO);
    if (this->scalar == NULL)
    {  
        /* can't register card register address space */
    	printf("configHy8522: Unable to register IO space\n");
        return;
    }

    /* map the memory address spaces  */
    this->convMemoryBase = (USHORT *) ipmBaseAddr(carrierNum, ipSlotNum, ipac_addrMem);
    if (this->convMemoryBase == NULL)
    {  
        /* can't register card Memory address space */
     	printf("configHy8417: Unable to register memory space\n");
        return;
    }
    
   /* Compare the HY8522 prom header, manufacturer's ID and card model
     * number with the expected values. */
    USHORT *tmp = (USHORT*)this->scalar->prom.vita4;
  	idmem =  *tmp;
  	if((idmem == 0x5649) || (idmem == 0x4956));
  	{
  		lstr[0] = 0x56; lstr[1] = 0x49;
  	}
  	idmem =  *((USHORT*)this->scalar->prom.vita4 + 1);
  	if((idmem == 0x5441) || (idmem == 0x4154));
  	{	
  		lstr[2] = 0x54; lstr[3] = 0x41;
  	}
  	idmem =  *((USHORT*)this->scalar->prom.vita4 + 2);
  	if((idmem == 0x3420) || (idmem == 0x2034));
  	{	
  		lstr[4] = 0x20; lstr[5] = 0x34;
  	}
  	lstr[6] = 0;

    strok = memcmp(lstr, VITA4, sizeof(this->scalar->prom.vita4)) == 0;
    ishytec = this->scalar->prom.hytecIdHigh == HYTECID_HI  &&
                  this->scalar->prom.hytecIdLow == HYTECID_LO;
    ismodel = this->scalar->prom.model == PROM_MODEL;

	/* check if we have been asked to be more chatty */
    /* start with some basic information */
    printf("PROM: Contents at offset - %p\n", &this->scalar->prom);
    printf("PROM: Prom header string - '%6s'\n", lstr);
    printf("PROM: Manufacturer's ID  - 0x%04x%04x %s\n",
        (int)this->scalar->prom.hytecIdHigh, (int)this->scalar->prom.hytecIdLow, 
        ishytec ? hytecstr : "(Unknown)");
    printf("\n");
    /* output a bit more about this specific model and card */
    printf("PROM: Model #/revision   - %#04hx rev. %#04hx\n",
        (int)this->scalar->prom.model, (int)this->scalar->prom.revision);

	ver = (int)this->scalar->prom.revision;
	ver = ver/4096*1000 + (ver%4096)/256*100 + ((ver%4096)%256)/16*10 + ((ver%4096)%256)%16;
	setIntegerParam(hy8522FWVersion_, ver);

    printf("PROM: Serial number      - %hu\n", (int)this->scalar->prom.serialNumber);
    printf("PROM: Driver IDs         - %#04hx, %#04hx\n",
        (int)this->scalar->prom.driverIdHigh, (int)this->scalar->prom.driverIdLow);
    printf("PROM: Device flags       - %#04hx\n", (int)this->scalar->prom.flags);
    printf("PROM: Bytes used         - %#04hx (%hd)\n",
        (int)this->scalar->prom.byteCount, (int)this->scalar->prom.byteCount);
    printf("\n");

    /* check if everything passed okay */
    if (!strok)
        printf("PROM: INVALID PROM HEADER; EXPECTED '%s'\n", VITA4);
    if (!ishytec)
        printf("PROM: UNSUPPORTED MANUFACTURER ID; EXPECTED 0x%04x%04x, %s\n",
            HYTECID_HI, HYTECID_LO, hytecstr);
    if (!ismodel)
	{
        printf("PROM: UNSUPPORTED MODEL NUMBER; EXPECTED %#04hx\n",
            PROM_MODEL);
	}
   
    /* set driver version */
	setIntegerParam(hy8522DriverVersion_, DRIVER_VERSION);
	setIntegerParam(hy8522Support_, DRIVER_SUPPORT);
	
    /* clear asyn message sent counters */
    this->messagesSent = 0;
    this->messagesFailed = 0;

    this->rebooting_ = 0;
    
	// tell listeners the changes
	callParamCallbacks();
   	epicsAtExit(rebootCallbackC, this);
}

void Hy8522Asyn::setupInterrupt()
{
    /* set up interrupt, work queue and mutex stuff */
    this->scalar->reg.vector = this->vectorNum_;            									/* writes vector register offset 0x08 */
    driverTable[numCards] = this;
    numCards++;
	if (ipmIntConnect(this->carrierNum_, this->ipSlotNum_, this->vectorNum_, INTSRC, numCards-1) < 0) {
      printf("configHy8522: failure connecting to interrupt\n");
      return;
    }

    ipmIrqCmd(this->carrierNum_, this->ipSlotNum_, 0, ipac_irqEnable);

    this->intMsgQId = epicsMessageQueueCreate(MAX_MESSAGES, 
    					(MAX_CHANNELS + 2)*sizeof(int));
    if (epicsThreadCreate("Hy8522intQueuedTask",
    					epicsThreadPriorityLow,
                        epicsThreadGetStackSize(epicsThreadStackMedium),
                        (EPICSTHREADFUNC)intQueuedTaskC,
                        (void *)this) == NULL)
	{	
      	printf("configHy8522: failure creating message queue.\n");
      	return;
	}	
}

int Hy8522Asyn::Hy8522InitCoin0_7(int coin0, int coin1, int coin2, int coin3, 
		int coin4, int coin5, int coin6, int coin7)
{
	this->scalar->reg.coin[0] = coin0;
	this->scalar->reg.coin[1] = coin1;
	this->scalar->reg.coin[2] = coin2;
	this->scalar->reg.coin[3] = coin3;
	this->scalar->reg.coin[4] = coin4;
	this->scalar->reg.coin[5] = coin5;
	this->scalar->reg.coin[6] = coin6;
	this->scalar->reg.coin[7] = coin7;
	return(asynSuccess);
}

int Hy8522Asyn::Hy8522InitCoin8_15(int coin8, int coin9, int coin10, int coin11, 
		int coin12, int coin13, int coin14, int coin15)
{	
	this->scalar->reg.coin[8] = coin8;
	this->scalar->reg.coin[9] = coin9;
	this->scalar->reg.coin[10] = coin10;
	this->scalar->reg.coin[11] = coin11;
	this->scalar->reg.coin[12] = coin12;
	this->scalar->reg.coin[13] = coin13;
	this->scalar->reg.coin[14] = coin14;
	this->scalar->reg.coin[15] = coin15;
	return(asynSuccess);	
}

/** Histogram configuration command, called directly or from iocsh */
/* 
 * Where: -
 *     gateIntv = When extadvance setting is 0 , this is the gate interval as below: 
 *     					If timeunit = 0, i.e. 1ms, this defines the interval in mini-second 
 *     			  		If timeunit = 1, i.e. 100us, this defines the interval in 100us steps. 
 *     			  When extadvance = 1, it is number of pulses to advance the memory location
 *     timeunit = Define interval time unit. 0 = 1ms, 1 = 100us
 *     nocycles	= Define the number of cycles. If it is set to 0, the scaler doesn't do anything.
 *      nogates = Define the number of gates (bins), i.e. How many gates in a cycle. Each gate 
 *      				stores counting of pulses in one memory location. Next cycle adds up to 
 *      				the same locatoin to form histogram.
 *         mask = bit mask to enable/disable channels. 0 = enable, 1 = disable.
 *   extadvance = Memory advance control. 0 = advanced by interval timer, 1 = advanced by external hardware pulses 
 *   accumulate = Define the returning waveform of the histogram. =0 return normal histogram. = 1 return accumulated 
 *				  		waveform -- each bin also accumulates all previous bins.
 */
int Hy8522Asyn::Hy8522HistogramInit(int gateIntv, int timeunit, int nocycles, int nogates, int mask, int extadvance, int accumulate)
{
	setIntegerParam(hy8522Mode_, hy8522histogramnormal);											//normal histogram mode
	
    //change firmware if mode is different from previous 
   	if(this->scalar->reg.function & MODE_PRESET)													//if it is preset scaler mdoe
   		this->scalar->reg.function = MODE_SET;														//set mode to change it    	

    epicsThreadSleep ( 1.000000 );																	//wait for a second to allow it to finish
    
    setupInterrupt();
    
	setIntegerParam(hy8522GateIntv_, gateIntv);
	setIntegerParam(hy8522TimeUnit_, timeunit);
	setIntegerParam(hy8522NoCycles_, nocycles);
	setIntegerParam(hy8522NoGates_, nogates);
	setIntegerParam(hy8522DiableInput_, mask);
	setIntegerParam(hy8522ExtAdvance_, extadvance);
	setIntegerParam(hy8522Addup_, accumulate);															//initial mode to either normal or addup histogram.

	/*******************************************************************/
	/************* Working out real stuffs for ADC *********************/
	/*******************************************************************/
	setupScaler();

	// tell listeners the changes
	callParamCallbacks();
	return(asynSuccess);
}

/** Coincidence histogram configuration command, called directly or from iocsh */
/* 
 * Where: -
 *     gateIntv = When extadvance setting is 0 , this is the gate interval as below: 
 *     					If timeunit = 0, i.e. 1ms, this defines the interval in mini-second 
 *     			  		If timeunit = 1, i.e. 100us, this defines the interval in 100us steps. 
 *     			  When extadvance = 1, it is number of pulses to advance the memory location
 *     timeunit = Define interval time unit. 0 = 1ms, 1 = 100us
 *     nocycles	= Define the number of cycles. If it is set to 0, the scaler doesn't do anything.
 *      nogates = Define the number of gates (bins), i.e. How many gates in a cycle. Each gate 
 *      				stores counting of pulses in one memory location. Next cycle adds up to 
 *      				the same locatoin to form histogram.
 *         mask = bit mask to enable/disable channels. 0 = enable, 1 = disable.
 *   extadvance = Memory advance control. 0 = advanced by interval timer, 1 = advanced by external hardware pulses 
 *   accumulate = Define the returning waveform of the histogram. =0 return normal histogram. = 1 return accumulated 
 *				  		waveform -- each bin also accumulates all previous bins.
 */
int Hy8522Asyn::Hy8522CoinInit(int gateIntv, int timeunit, int nocycles, int nogates, int mask, int extadvance, int accumulate)
{
	setIntegerParam(hy8522Mode_, hy8522histogramcoincidence);										//coincidence histogram mode
	
    //change firmware if mode is different from previous 
   	if(this->scalar->reg.function & MODE_PRESET)													//if it is preset scaler mdoe
   		this->scalar->reg.function = MODE_SET;														//set mode to change it    	

    epicsThreadSleep ( 1.000000 );																	//wait for a second to allow it to finish
    
    setupInterrupt();
    
	setIntegerParam(hy8522GateIntv_, gateIntv);
	setIntegerParam(hy8522TimeUnit_, timeunit);
	setIntegerParam(hy8522NoCycles_, nocycles);
	setIntegerParam(hy8522NoGates_, nogates);
	setIntegerParam(hy8522DiableInput_, mask);
	setIntegerParam(hy8522ExtAdvance_, extadvance);
	setIntegerParam(hy8522Addup_, accumulate);															//initial mode to either normal or addup histogram.

	/*******************************************************************/
	/************* Working out real stuffs for ADC *********************/
	/*******************************************************************/
	setupScaler();

	// tell listeners the changes
	callParamCallbacks();
	return(asynSuccess);
}

/** Straight scaler configuration command, called directly or from iocsh */
/* 
 * Where: -
 *     gateIntv = Define gate interval as below: 
 *     					If timeunit = 0, i.e. 1ms, this defines the interval in mini-second 
 *     			  		If timeunit = 1, i.e. 100us, this defines the interval in 100us steps. 
 *     timeunit = Define interval time unit. 0 = 1ms, 1 = 100us
 *     nocycles	= Define the number of trigger cycles. Up to 32K.
 *         mask = bit mask to enable/disable channels. 0 = enable, 1 = disable.
 */
int Hy8522Asyn::Hy8522StraightInit(int gateIntv, int timeunit, int nocycles, int mask)
{
	setIntegerParam(hy8522Mode_, hy8522straightscaler);												//straight scaler mode
	
    //change firmware if mode is different from previous 
   	if(this->scalar->reg.function & MODE_PRESET)											//if it is preset scaler mdoe
   		this->scalar->reg.function = MODE_SET;												//set mode to change it    	

    epicsThreadSleep ( 1.000000 );																//wait for a second to allow it to finish
    
    setupInterrupt();
    
	setIntegerParam(hy8522GateIntv_, gateIntv);
	setIntegerParam(hy8522TimeUnit_, timeunit);
	setIntegerParam(hy8522NoCycles_, nocycles);
	setIntegerParam(hy8522DiableInput_, mask);

	/*******************************************************************/
	/************* Working out real stuffs for ADC *********************/
	/*******************************************************************/
	setupScaler();

	// tell listeners the changes
	callParamCallbacks();
	return(asynSuccess);
}

/** Preset scaler configuration command, called directly or from iocsh */
/* 
 * Where: -
 *         mask = Define ARM bit. Set to 1 to ARM the correspondent channel
 */
int Hy8522Asyn::Hy8522PresetInit(int mask)
{
	setIntegerParam(hy8522Mode_, hy8522presetscaler);												//preset scaler mode
	
	if(!(this->scalar->reg.function & MODE_PRESET))											//if it is not preset scaler mdoe
		this->scalar->reg.function = MODE_SET;												//set mode to change it 

	epicsThreadSleep ( 1.000000 );																//wait for a second to allow it to finish
    
    setupInterrupt();
	
	setIntegerParam(hy8522DiableInput_, mask);

	/*******************************************************************/
	/************* Working out real stuffs for ADC *********************/
	/*******************************************************************/
	setupScaler();

	// tell listeners the changes
	callParamCallbacks();
	return(asynSuccess);
}


/*********************************************************/
/*                                                      */
/* Asyn interface Function implementations             */
/*                                                    */
/*****************************************************/

/* asynCommon methods */
void Hy8522Asyn::report(FILE *fp, int details)
{
	int i, mode, unit, hwtrigger, hwreset, gateintv, nocycles, nogates, mask, fwver, swver, extadv, freq, source, restart;
	int count0, count1;

    fprintf(fp, "Port: %s, carrier slot %d slot %d, base address=%p, membase address=%p\n", 
            this->portName, this->carrierNum_, this->ipSlotNum_, (void *)this->scalar, (void *)this->convMemoryBase);
    if (details == 1) 
    {
		getIntegerParam(hy8522Mode_, &mode);
		getIntegerParam(hy8522TimeUnit_, &unit);
		getIntegerParam(hy8522HWTrigger_, &hwtrigger);
		getIntegerParam(hy8522HWReset_, &hwreset);
		getIntegerParam(hy8522GateIntv_, &gateintv);
		getIntegerParam(hy8522NoCycles_, &nocycles);
		getIntegerParam(hy8522NoGates_, &nogates);
		getIntegerParam(hy8522DiableInput_, &mask);
		getIntegerParam(hy8522FWVersion_, &fwver);
		getIntegerParam(hy8522DriverVersion_, &swver);
		getIntegerParam(hy8522ExtAdvance_, &extadv);
		getIntegerParam(hy8522Frequency_, &freq);
		getIntegerParam(hy8522Source_, &source);
		getIntegerParam(hy8522Restart_, &restart);
		
        fprintf(fp, "  Current csr=%x\n", this->scalar->reg.csr);
        fprintf(fp, "  INT Vector=%x\n", this->vectorNum_);
        fprintf(fp, "  Mode=%s\n", mode==0?"histogram normal":(mode==1?"histogram coincidence":(mode==2?"straight scalar":"preset scalar")));
        fprintf(fp, "  Time unit=%s\n", unit==0?"1ms":"100us");
        fprintf(fp, "  Allow HW trigger=%s\n", hwtrigger==0?"No":"Yes");
        fprintf(fp, "  Allow HW reset=%s\n", hwreset==0?"No":"Yes");
        fprintf(fp, "  Gate interval=%d\n", gateintv);
        fprintf(fp, "  Number of cycles=%d\n", nocycles);
        fprintf(fp, "  Number of gates per cycle=%d\n", nogates);
        fprintf(fp, "  Channle mask=0x%x\n", mask);
        fprintf(fp, "  Ext Adcance:%s\n", (extadv == 0) ? "No": "Yes");
        fprintf(fp, "  Frequency = %d\n", freq);
        fprintf(fp, "  Source = %d\n", source);
        fprintf(fp, "  function=%x\n", this->scalar->reg.function);
        fprintf(fp, "  restart mode=%d\n", restart);
        fprintf(fp, "  Firmware Version=%d\n", fwver);
        fprintf(fp, "  Driver Version=%d\n", swver);
        fprintf(fp, "  messages sent OK=%d; send failed (queue full)=%d\n",
                this->messagesSent, this->messagesFailed);       
    }
    if ((details == 1) && (mode == hy8522presetscaler))
    {
        fprintf(fp, "  Preset hit=%x\n", this->scalar->reg.overflow);
        fprintf(fp, "  Preset mask=%x\n", this->scalar->reg.triggerno);
        fprintf(fp, "  Preset ARM=%x\n", this->scalar->reg.disablechl);
    }
    if ((details == 1) && (mode == hy8522presetscaler))
    {
    	for(i=0; i<MAX_CHANNELS; i++)
    	{
    		count0 = this->scalar->reg.reg[2*i];
    		count1 = this->scalar->reg.reg[2*i +1] * 65536;
            fprintf(fp, "  Counter%d=%d\n", i, count0+count1);
    	}
    	fprintf(fp, "\n");
    }
    if ((details == 1) && (mode == hy8522histogramcoincidence))
    {
    	for(i=0; i<MAX_CHANNELS; i++)
    	{
            fprintf(fp, "  Coincidence%d=%x\n", i, this->scalar->reg.coin[i]);
    	}
    	fprintf(fp, "\n");
    }
    // Call the base class method
    asynPortDriver::report(fp, details);
}


/*****         asynInt32 methods       ******/
asynStatus Hy8522Asyn::writeInt32(asynUser *pasynUser, epicsInt32 value)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
	int function = pasynUser->reason;
	status = this->getAddress(pasynUser, &channel);
	int i, mode;
    int presetmask, imask;
    unsigned short* gptr;
    unsigned int* prptr;
	
    scalerAsynPvt *pPvt_curr = (scalerAsynPvt *)pasynUser->userPvt;						/* current user asynScaler device pointer */
	scalerRecord *psr_curr = pPvt_curr->psr;											/* current user scaler record */
	
	getIntegerParam(hy8522Mode_, &mode);

/* decide what to do depending on which command we received */
	if (function == hy8522Coin_)
		status = setCoin( channel, value );
	else if (function == hy8522HWTrigger_)
		status = setEnableHWTrigger( value );
	else if (function == hy8522HWReset_)
		status = setEnableHWReset( value );
	else if (function == hy8522SoftTrigger_)
		status = softwareTrigger();
	else if (function == hy8522GateIntv_)
		status = setGateInterval( value );
	else if (function == hy8522NoCycles_)
		status = setNoOfCycles( value );
	else if (function == hy8522NoGates_)
		status = setNoOfGates( value );
	else if (function == hy8522DiableInput_)
		status = setMask( value );
	else if (function == hy8522TimeUnit_)
		status = setTimeUint( value );
	else if (function == hy8522ExtAdvance_)
		status = setExtAdvance( value );
	else if (function == hy8522Addup_)
		status = setAddupMode( value );
	else if (function == hy8522Frequency_)
		status = setFrequency( value );
	else if (function == hy8522Source_)
		status = setSource( value );
	else if(function == hy8522Restart_)
		status = restart( value );

	
	// Scaler commands
	else if ((function == scalerPresets_)  && (mode == hy8522presetscaler))
	{
		if(psr_curr != this->psr) 														//if the asyn user is not from scaler record 
		{
			if(channel != 0)															//channel 0 is used as clock, so cannot be changed. It is controlled by scaler record
			{
				//post it back to scaler record
				gptr=&(this->psr->g1);
				prptr=(unsigned int*)(&(this->psr->pr1));
				gptr[channel] = 1;
				db_post_events(this->psr,&(gptr[channel]),DBE_VALUE);
				prptr[channel] = value;
				db_post_events(this->psr,&(prptr[channel]),DBE_VALUE);					//post it back		
		
				setIntegerParam(channel, scalerPresets_, value);						//save value
			}
		}else																			//if the asyn user is scaler record
		{
			gptr=&(this->psr->g1);
			prptr=(unsigned int*)(&(this->psr->pr1));
			if(channel == 0)															//this is the clock channel
			{
				if((this->psr->cont == 1) && (this->psr->cnt != 1))						//autocount and 'Count' is not in action.
					value = (epicsInt32) (this->psr->freq * this->psr->tp1);
				else																	//OneShot
					value = (epicsInt32) (this->psr->freq * this->psr->tp);

				prptr[channel] = value;													//set it back to the scaler record
				db_post_events(this->psr,&(prptr[channel]),DBE_VALUE);					//post it back
			}else
			{
				if((value == 0)	&& (gptr[channel]))										//if value = 0 and it is a preset counter, we give it a initial value of 1000 counts.
				{
					value = 1000;
					prptr[channel] = value;												//set it back to the scaler record
					db_post_events(this->psr,&(prptr[channel]),DBE_VALUE);				//post it back
				}
			}
			setIntegerParam(channel, scalerPresets_, value);							//save value				
		}
		callParamCallbacks(channel);
	}
	else if ((function == scalerReset_)  && (mode == hy8522presetscaler))
	{
		/* Reset scaler */
	    resetScaler();
	    setIntegerParam(scalerDone_, 0);												//clear the done bit
		callParamCallbacks();
	    /* Clear all of the presets and counts*/
	    for (i=0; i<MAX_CHANNELS; i++) {
	    	setIntegerParam(i, scalerPresets_, 0);
		    callParamCallbacks(i);
	    }
	}
	else if ((function == scalerArm_)  && (mode == hy8522presetscaler))
	{
	    /* Arm or disarm scaler */
	    if (value != 0) 
		{
	    	if(this->psr->cnt == 1)														//OneShot
	    	{
		    	//figure out which one is preset which one is not
				gptr=&(this->psr->g1);
				prptr=&(this->psr->pr1);
				presetmask=0;												
				imask=1;

				for (i=0;i<MAX_CHANNELS;i++)
				{
					if (*gptr)															// Pre-set scaler 
						presetmask|=imask;
	  
					imask=imask<<1;
				    gptr++;
				}

				this->presetmask=presetmask;	
	    	}
	    	else																		//auto count
				this->presetmask=1;														//only channel 0 is the preset for timing control	
	
    	   	setScalerPresets();
	      	startScalers();
	    } else 
		{
	    	stopScalers();
		    setIntegerParam(scalerDone_, 0);											//stop is not the same as 'finish'. So should set done to 0!
	    }
		callParamCallbacks();
	}
	
	return( status );
}

asynStatus Hy8522Asyn::readInt32(asynUser *pasynUser, epicsInt32 *value)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
	int function = pasynUser->reason;
	int mode, triggerState, f, done;
	USHORT counterL, counterH;
    int chans;
   
    if (this->rebooting_) epicsThreadSuspendSelf();
  	
	getIntegerParam(hy8522Mode_, &mode);
	getIntegerParam(hy8522TriggerStatus_, &triggerState);

	this->getAddress(pasynUser, &channel);

	if(function == hy8522Data_)
	{
		counterL = this->scalar->reg.reg[channel*2];
		counterH = this->scalar->reg.reg[channel*2+1];
		*value = (epicsInt32) counterL | (epicsInt32) counterH<<16;
/* Testing */
/*
if(channel == 0)
{
int ver;
ver = this->scalar->reg.csr;
printf("chn=%d vl=%x vh=%x cv=%x csr=%x\n", channel, counterL, counterH<<16, *value, ver);
}
*/
	}
	else if(function == hy8522TriggerStatus_)
	{
       	*value = triggerState;
	}
	else if(function == hy8522NoTriggers_)
	{
       	*value = this->scalar->reg.triggerno;
	}
	else if(function == hy8522MemPointer_)
	{
		if((mode == hy8522histogramnormal) || (mode == hy8522histogramcoincidence) || (mode == hy8522straightscaler)) {
			//this return number of items saved in memory per channel
			*value = (epicsInt32) ((this->scalar->reg.pointerh * 65536 + this->scalar->reg.pointerl) / 32);
			setIntegerParam(hy8522MemPointer_, *value);
		}
	}
	
	//scaler read
	else if ((function == scalerReadSingle_) && (mode == hy8522presetscaler)) {
		readScalers();
	    /* Read a single scaler channel */
	    *value = basic[channel][0];
	}
	else if ((function == scalerDone_)  && (mode == hy8522presetscaler)) {
		getIntegerParam(scalerDone_, &done);
	    *value = done;
	}
	else if((function == scalerChannels_) && (mode == hy8522presetscaler)) {									//asking for channel numbers. This is called in the init of devScalerAsyn.c device support module
		getIntegerParam(scalerChannels_, &chans);
		*value = (epicsInt32) chans;
		
		//scaler record stuff
		this->pPvt = (scalerAsynPvt *)pasynUser->userPvt;														//first time called by scaler record, let's remember it
		this->psr = pPvt->psr;
	    	
		//resolve internal frequency. At this point, the scaler record has populated its fields.
		if(this->psr->freq==25000000)
			setIntegerParam(hy8522Frequency_, 0);																//25MHz by default
		else if(this->psr->freq==50000000)
			setIntegerParam(hy8522Frequency_, 1);																//50MHz by default
		else if(this->psr->freq==100000000)
			setIntegerParam(hy8522Frequency_, 2);																//100MHz by default
		else if(this->psr->freq==200000000)
			setIntegerParam(hy8522Frequency_, 3);																//200MHz by default
		getIntegerParam(hy8522Frequency_, &f);
		setFrequency( f );																						//set frequency
	}
	else
	{
		status = asynPortDriver::readInt32(pasynUser, value);
	}

	return( status );
}

/* asynInt32Array methods */
asynStatus Hy8522Asyn::readInt32Array(asynUser *pasynUser, epicsInt32 *data, size_t nelements, size_t *nactual)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
    int i, numberofpoint;
	int function = pasynUser->reason;
	int mode, nocycles, state;
	size_t j;

    if (this->rebooting_) epicsThreadSuspendSelf();
	
	getIntegerParam(hy8522Mode_, &mode);
	getIntegerParam(hy8522NoCycles_, &nocycles);
	getIntegerParam(hy8522NoGates_, &numberofpoint);
	getIntegerParam(hy8522TriggerStatus_, &state);								//get the current trigger state
	this->getAddress(pasynUser, &channel);
	
    if(function == hy8522Data_)
    {
		this->scalar->reg.csr &= ~CSR_ETM;																		//for straight scaler mode to fetch counting numbers
		if (mode == hy8522straightscaler)
		{
			numberofpoint = (this->scalar->reg.pointerh * 65536 + this->scalar->reg.pointerl) / 32;				//get current pointer, 2 words, 16 channels
			if (numberofpoint == 0)
			{
				if (this->isdone_) numberofpoint = nocycles;														//current pointer =0, check if it is the number of cycles that has been reached
			}
			else
			{
				this->isdone_ = false;																			//started again, clear finished flag
			}

			if (numberofpoint == 0)																				//this is to avoid it remembers the last values
			{
				for (j = 0; j < nelements; j++) *(data + j) = 0;
			}
			else
			{
				for (i = 0; i < numberofpoint; i++)
				{
					if (i<(int)nelements)
						*(data + i) = (this->convMemoryBase + i * 32)[channel * 2] +
						(this->convMemoryBase + i * 32)[channel * 2 + 1] * 65536;
				}
			}
		}
		*nactual = numberofpoint >(int) nelements ? nelements : numberofpoint;       							//return the real number of element                                        	

		//capture hardware triggerring though a bit late or even not get here
		if((this->scalar->reg.csr & CSR_ET) && 
			((this->scalar->reg.pointerh * 65536 + this->scalar->reg.pointerl) != 0) &&
			(state != 2))
			setIntegerParam(hy8522TriggerStatus_, TRIGGERRED_COLLECTING);									//collecting started. Set trigger status to collecting
	}
	else if ((function == scalerRead_)  && (mode == hy8522presetscaler)) {										//for preset scaler mode to read single channel
	    readScalers();
	    /* Read a single scaler channel */
	    for (i=0; (i<(int)nelements && i<(int)MAX_CHANNELS); i++) {
	    	data[i] = basic[i][0];
	    }
	    for (i=(int)MAX_CHANNELS; i<(int)nelements; i++) {
	    	data[i] = 0;
	    }
	    *nactual = nelements;
	}

	status = asynSuccess;
	return( status );
}

/* asynFloat64Array methods */
asynStatus Hy8522Asyn::readFloat64Array(asynUser *pasynUser, epicsFloat64 *data, size_t nelements, size_t *nactual)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
    int i, numberofpoint, startPoint = 0;
	int function = pasynUser->reason;
	int mode, addup, state;
	USHORT v0, v1, v2, v3;
	long long v = 0, sum = 0;
	
    if (this->rebooting_) epicsThreadSuspendSelf();
	
	getIntegerParam(hy8522Mode_, &mode);
	getIntegerParam(hy8522Addup_, &addup);
	getIntegerParam(hy8522NoGates_, &numberofpoint);
	getIntegerParam(hy8522TriggerStatus_, &state);								//get the current trigger state
	this->getAddress(pasynUser, &channel);
	
    if(function == hy8522Data_)
    {
    	if((mode == hy8522histogramnormal) || (mode == hy8522histogramcoincidence))
    	{
			this->scalar->reg.csr &= ~CSR_ETM;																//for histogram mode to fetch bin countings
			startPoint = 0;
			sum = 0;
			for (i = 0; i<(int)nelements-1; i++)
    		{
				if (i < numberofpoint)
				{

					v0 = (this->convMemoryBase + channel * CHANNELBANKSIZE)[startPoint];
					v1 = (this->convMemoryBase + channel * CHANNELBANKSIZE)[startPoint + 1];
					v2 = (this->convMemoryBase + channel * CHANNELBANKSIZE)[startPoint + 2];
					v3 = (this->convMemoryBase + channel * CHANNELBANKSIZE)[startPoint + 3];
					v = ((long long)v3 << 48) + ((long long)v2 << 32) + ((long long)v1 << 16) + v0;
					if (addup == 1)
					{
						sum += v;
						*(data + i) = (double)sum;
					}
					else
						*(data + i) = (double)v;
					startPoint += 4;
				}else
					*(data + i) = 0;
			}
/*for testing */
//printf("n=%d ch=%d, start=%d, csr=%x v=%d\n", numberofpoint, channel, startPoint, this->scalar->reg.csr, v);
//printf("\n");   		

			if(addup == 0)
			{
				this->scalar->reg.csr |= CSR_ETM;																//fetch totaliser data
				v0 = (this->convMemoryBase + channel * 4)[0];
				v1 = (this->convMemoryBase + channel * 4)[1];
				v2 = (this->convMemoryBase + channel * 4)[2];
				v3 = (this->convMemoryBase + channel * 4)[3];
				v = ((long long)v3<<48) + ((long long)v2<<32) + ((long long)v1<<16) + v0;
    			*(data + i) = (double) v;
			}
			else
				*(data + i) = (double) sum;
			numberofpoint++;
			
			//capture hardware triggerring though a bit late or even not get here
			if((this->scalar->reg.csr & CSR_ET) && 
				((this->scalar->reg.pointerh * 65536 + this->scalar->reg.pointerl) != 0) &&
				(state != 2))
				setIntegerParam(hy8522TriggerStatus_, TRIGGERRED_COLLECTING);									//collecting started. Set trigger status to collecting
    	}
		*nactual = numberofpoint >(int)nelements ? nelements : numberofpoint;									//return the real number of element                                        	
	}

	status = asynSuccess;
	return( status );
}

/*********************************************************/
/*                                                      */
/* Private Sector                                      */
/*                                                    */
/*****************************************************/

void Hy8522Asyn::setupScaler()
{
	int mode, unit, hwtrigger, hwreset, gateintv, nocycles, nogates, mask, adv, freq, source;

	if (this->rebooting_)  epicsThreadSuspendSelf();

	this->scalar->reg.csr = 0;
	resetScaler();
	this->scalar->reg.csr |= CSR_IE;  																			//reset sacler and enable interrupt   
//	epicsThreadSleep ( 1.000000 );
	
	getIntegerParam(hy8522Mode_, &mode);
	getIntegerParam(hy8522TimeUnit_, &unit);
	getIntegerParam(hy8522HWTrigger_, &hwtrigger);
	getIntegerParam(hy8522HWReset_, &hwreset);
	getIntegerParam(hy8522GateIntv_, &gateintv);
	getIntegerParam(hy8522NoCycles_, &nocycles);
	getIntegerParam(hy8522NoGates_, &nogates);
	getIntegerParam(hy8522DiableInput_, &mask);
	getIntegerParam(hy8522ExtAdvance_, &adv);
	getIntegerParam(hy8522Frequency_, &freq);
	getIntegerParam(hy8522Source_, &source);
	setEnableHWTrigger( hwtrigger );	          
	setEnableHWReset( hwreset );
	
	if(mode != hy8522presetscaler)
	{
		//these only need when it is histogram or straight scaler mode
		setGateInterval( gateintv );
		setNoOfCycles( nocycles );
		setNoOfGates(nogates);																					//number of gates 
		setExtAdvance( adv );		
	}
	setMask(mask);																								//mask
	setTimeUint(unit);																							//time unit
	setFrequency( freq );
	setSource(source);																							//source
	setMode( mode );
}

asynStatus Hy8522Asyn::setMode( int mode  )
{
	USHORT csr;
	if (this->rebooting_)  epicsThreadSuspendSelf();

	if(mode < 0 || mode > 3) return(asynError);
	
	this->scalar->reg.csr &= ~CSR_MODE_UNUSED;																	//clear mdoe first
	if ((mode == hy8522histogramnormal) || (mode == hy8522histogramcoincidence))								//if it is histogram mode
	{
    	if(mode == hy8522histogramcoincidence)																	//coincidence
    		this->scalar->reg.csr |= CSR_EC;
    	else
    		this->scalar->reg.csr &= ~CSR_EC;																	//normal
	}else if(mode == hy8522straightscaler)
	{
		csr = CSR_MODE_STRAIGHT;
		this->scalar->reg.csr |= csr; 			   																//set mode		
	}else if(mode == hy8522presetscaler)
	{
		//csr = CSR_MODE_PRESET;
		//this->scalar->reg.csr |= csr; 			   															//set mode				
	}

	setIntegerParam(hy8522Mode_, mode);
	callParamCallbacks();
	return(asynSuccess);
}

asynStatus Hy8522Asyn::setEnableHWTrigger( int hwtrigger )
{
	if (this->rebooting_)  epicsThreadSuspendSelf();

	if(hwtrigger < 0 || hwtrigger > 1) return(asynError);
	
	if (hwtrigger)
    	this->scalar->reg.csr |= CSR_ET;																		//allow hardware trigger
    else
    	this->scalar->reg.csr &= ~CSR_ET;    																	//no hardware trigger

	setIntegerParam(hy8522HWTrigger_, hwtrigger);
	callParamCallbacks();
	return(asynSuccess);
}

asynStatus Hy8522Asyn::softwareTrigger()
{
	if (this->rebooting_)  epicsThreadSuspendSelf();

    this->scalar->reg.csr &= ~CSR_ST;                           												//clear software trigger latch 
    this->scalar->reg.csr |= CSR_ST;                            												//software trigger to kick off the acquisition
    this->scalar->reg.csr &= ~CSR_ST;                           												//clear software latch to enable hardware trigger afterwards
	setIntegerParam(hy8522TriggerStatus_, TRIGGERRED_COLLECTING);												//collecting started. Set trigger status to collecting
    return( asynSuccess );
}

asynStatus Hy8522Asyn::setEnableHWReset( int hwreset )
{
	if (this->rebooting_)  epicsThreadSuspendSelf();

	if(hwreset < 0 || hwreset > 1) return(asynError);
	
	if (hwreset)
    	this->scalar->reg.csr |= CSR_ER;																		//allow hardware reset
    else
    	this->scalar->reg.csr &= ~CSR_ER;    																	//no hardware reset

	setIntegerParam(hy8522HWReset_, hwreset);
	callParamCallbacks();
	return(asynSuccess);
}

asynStatus Hy8522Asyn::setGateInterval( int gateintv )
{
	if (this->rebooting_)  epicsThreadSuspendSelf();

   	this->scalar->reg.gateintv = gateintv;																		//gate interval

	setIntegerParam(hy8522GateIntv_, gateintv);
	callParamCallbacks();
	return(asynSuccess);
}

asynStatus Hy8522Asyn::setNoOfCycles(int nocycles)
{
	if (this->rebooting_)  epicsThreadSuspendSelf();
	
  	this->scalar->reg.cycles = nocycles;																		//number of cycles

	setIntegerParam(hy8522NoCycles_, nocycles);
	callParamCallbacks();
	return(asynSuccess);
}

asynStatus Hy8522Asyn::setNoOfGates(int nogates)
{
	if (this->rebooting_)  epicsThreadSuspendSelf();

	if(nogates < 0 || nogates > 0x4000) return(asynError);

   	this->scalar->reg.gates = nogates;																			//number of gates

	setIntegerParam(hy8522NoGates_, nogates);
	callParamCallbacks();
	return(asynSuccess);
}

asynStatus Hy8522Asyn::setMask(int mask)
{
	if (this->rebooting_)  epicsThreadSuspendSelf();

   	this->scalar->reg.disablechl = mask;																		//channel mask

	setIntegerParam(hy8522DiableInput_, mask);
	callParamCallbacks();
	return(asynSuccess);
}

asynStatus Hy8522Asyn::setTimeUint(int unit)
{
	if (this->rebooting_)  epicsThreadSuspendSelf();

	if(unit < 0 || unit > 1) return(asynError);
	
	if (unit)
    	this->scalar->reg.csr |= CSR_TM;																		//1ms
    else
    	this->scalar->reg.csr &= ~CSR_TM;    																	//100us

	setIntegerParam(hy8522TimeUnit_, unit);
	callParamCallbacks();
	return(asynSuccess);
}

asynStatus Hy8522Asyn::setCoin(int channel, int coin)
{	
	if (this->rebooting_)  epicsThreadSuspendSelf();

	if(channel < 0 || channel > 15) return(asynError);
	
  	this->scalar->reg.coin[channel] = coin;																		//set coincidence pattern

	return(asynSuccess);
}

asynStatus Hy8522Asyn::setExtAdvance( int extadv )
{
	if (this->rebooting_)  epicsThreadSuspendSelf();

	if(extadv != 0)
    	this->scalar->reg.csr |= CSR_EBA;																		//enable external advancing
	else
    	this->scalar->reg.csr &= ~CSR_EBA;																		//internal clock advancing

	setIntegerParam(hy8522ExtAdvance_, extadv);		
	return(asynSuccess);
}

asynStatus Hy8522Asyn::setAddupMode( int addup )
{
	if (this->rebooting_)  epicsThreadSuspendSelf();
	if(addup < 0 || addup > 1) return(asynError);
	
	setIntegerParam(hy8522Addup_, addup);
 	return(asynSuccess);
}

asynStatus Hy8522Asyn::setFrequency	( int freq )
{
	USHORT csr;
	if (this->rebooting_)  epicsThreadSuspendSelf();

	if(freq < 0 || freq > 3) return(asynError);
	
	setIntegerParam(hy8522Frequency_, freq);
	csr = freq << 13;
	this->scalar->reg.csr |= csr;																				//or bit 14,13
 	return(asynSuccess);
}

asynStatus Hy8522Asyn::setSource ( int source )
{
	if (this->rebooting_)  epicsThreadSuspendSelf();

	source = source & 0xFFFF;
	
	setIntegerParam(hy8522Source_, source);
	this->scalar->reg.intextclock = source;																		//set pulse source
 	return(asynSuccess);
}  	

asynStatus Hy8522Asyn::restart (int mode)
{
	int cycles = 0;
	if (this->rebooting_)  epicsThreadSuspendSelf();
	
	if(mode < 0 || mode > 3) return(asynError);																	//range check

	setIntegerParam(hy8522Restart_, mode);
	if(mode == hy8522carryonrestart)
	{
		this->scalar->reg.cycles = 0;																			// clear cycles   
		getIntegerParam(hy8522NoCycles_, &cycles);
		this->scalar->reg.cycles = cycles;																		//re-write cycles
	}else if((mode == hy8522cleanrestart) || (mode == hy8522cleanrestartatINT))
		resetScaler();
	
	setIntegerParam(hy8522TriggerStatus_, WAITING_TRIGGER);
	return(asynSuccess);
}  	


/*********************************************************/
/*                                                      */
/* Scaler record Routines       					   */
/*                                                    */
/*****************************************************/
asynStatus Hy8522Asyn::readScalers	()
{
	int i, bit = 1, pre_value;
	if (this->rebooting_)  epicsThreadSuspendSelf();

	for (i=0; i<MAX_CHANNELS; i++)
	{
    	if(this->presetmask & bit)
    	{
    		basic[i][0] = (ULONG) (this->scalar->reg.reg[i*2] +  this->scalar->reg.reg[i*2+1]*65536);
        	getIntegerParam(i, scalerPresets_, &pre_value);
        	if(basic[i][0] == 0)
    			basic[i][0] = pre_value;
    		else
    			basic[i][0] = pre_value - (MAX_COUNT - basic[i][0]);
    	}
    	else
    		basic[i][0] = (ULONG) (this->scalar->reg.reg[i*2] +  this->scalar->reg.reg[i*2+1]*65536);
    	bit = bit << 1;
	}

 	return(asynSuccess);
}

asynStatus Hy8522Asyn::resetScaler	()
{
	unsigned short csr = CSR_R, hd_tgr;
	if (this->rebooting_)  epicsThreadSuspendSelf();

	hd_tgr = this->scalar->reg.csr & CSR_ET;				//check if hardware trigger is allowed
	this->scalar->reg.csr &= ~CSR_ET;						//clear hardware trigger to stop immediately counting 
	this->scalar->reg.csr |= CSR_R;							//or bit 1 to reset the module
	while(csr) {csr = this->scalar->reg.csr & CSR_R;}
	if (hd_tgr != 0) this->scalar->reg.csr |= CSR_ET;		//if the hardware trigger is set before, restore to start next run
 	return(asynSuccess);
}

asynStatus Hy8522Asyn::setScalerPresets	()
{
	int i, pre_value, bit = 1;
	unsigned int starting_value = 0;
	if (this->rebooting_)  epicsThreadSuspendSelf();

    for (i=0; i<MAX_CHANNELS; i++) {
    	if(this->presetmask & bit)
    	{
        	getIntegerParam(i, scalerPresets_, &pre_value);
        	starting_value = MAX_COUNT - (unsigned int)pre_value;
        	this->scalar->reg.reg[i*2] = starting_value & 0xFFFF;
        	this->scalar->reg.reg[i*2+1] = starting_value >> 16;    		
    	}else
    	{
        	this->scalar->reg.reg[i*2] = 0;
        	this->scalar->reg.reg[i*2+1] = 0;    		    		
    	}
    	bit = bit << 1;
    }
	
 	return(asynSuccess);
}

asynStatus Hy8522Asyn::startScalers	()
{
	int mask;
	if (this->rebooting_)  epicsThreadSuspendSelf();

	getIntegerParam(hy8522DiableInput_, &mask);
  	this->scalar->reg.disablechl = (USHORT)mask;																//ARM channels
  	this->scalar->reg.triggerno = 0xFFFF;																		//allow everyone to overflow and stop
 	softwareTrigger();
 
 	return(asynSuccess);
}

asynStatus Hy8522Asyn::stopScalers	()
{
	if (this->rebooting_)  epicsThreadSuspendSelf();
  	this->scalar->reg.disablechl = 0;																			//dis-ARM channels
  	return(asynSuccess);
}

void Hy8522Asyn::rebootCallback()
{
    this->rebooting_ = 1;				                            											//flag we are rebooting
}


/*********************************************************/
/*                                                      */
/* External Wrapped Routines      					   */
/*                                                    */
/*****************************************************/

// These functions must have C linkage because they are called from other EPICS components
extern "C" {
static void rebootCallbackC(void *drvPvt)
{
    Hy8522Asyn *pPvt = (Hy8522Asyn *)drvPvt;
    pPvt->rebootCallback();
}

static void INTSRC(int cardNumber)
{
    
    Hy8522Asyn *pPvt = driverTable[cardNumber];
    pPvt->intSR();
}

static void intQueuedTaskC(void *drvPvt)
{
    Hy8522Asyn *pPvt = (Hy8522Asyn *)drvPvt;
    pPvt->intQueuedTask();	
}

}

/*********************************************************/
/*                                                      */
/* Interrupt and Queue Task and external Routines      */
/*                                                    */
/*****************************************************/
void Hy8522Asyn::intSR()
{
    int data[2];
    if (this->scalar->reg.csr & CSR_IRQ)     																	//Interrupt requested
    {
       this->scalar->reg.intdisable |= DIS_INT;                  												//disable interrupt
        /* Wake up queue task which calls callback routines */
         if (epicsMessageQueueTrySend(this->intMsgQId, data, sizeof(data)) == 0)
         	messagesSent++;
         else
         	messagesFailed++;
    }
//    callParamCallbacks();
}

void Hy8522Asyn::intQueuedTask()
{
    int i, chan, pointer;
    int data[2];
	int mode, noofpoints, cycles, restart, addup;
	USHORT v0, v1, v2, v3;
	long long v = 0, sum = 0;

    while(1) {
        /* Wait for event from interrupt routine */
        epicsMessageQueueReceive(intMsgQId, data, sizeof(data));
                 
     	getIntegerParam(hy8522Mode_, &mode);
		getIntegerParam(hy8522Addup_, &addup);
  
    	if(mode == hy8522straightscaler)																		//straight scaler
     	{
    		this->scalar->reg.csr &= ~CSR_ETM;																	//fetch counting numbers
         	getIntegerParam(hy8522NoCycles_, &noofpoints);
     		for (pointer=0; pointer<noofpoints; pointer++)
     		{
     			for (chan=0; chan<MAX_CHANNELS; chan++)
     			{
     				basic[chan][pointer] = (ULONG)((this->convMemoryBase + pointer * 32)[chan*2]) + 
        					(ULONG)((this->convMemoryBase + pointer * 32)[chan*2 + 1] * 65536);
     			}
     		}
     	}else if((mode == hy8522histogramnormal) || (mode == hy8522histogramcoincidence))						//histogram
     	{
         	getIntegerParam(hy8522NoGates_, &noofpoints);
 			for (chan=0; chan<MAX_CHANNELS; chan++)
     		{
 				this->scalar->reg.csr &= ~CSR_ETM;		
				pointer = 0;
				sum = 0;
     	     	for (i=0; i<noofpoints; i++)
    			{
     	     		v0 = (this->convMemoryBase + chan * CHANNELBANKSIZE)[pointer]; 
					v1 = (this->convMemoryBase + chan * CHANNELBANKSIZE)[pointer + 1];
					v2 = (this->convMemoryBase + chan * CHANNELBANKSIZE)[pointer + 2];
					v3 = (this->convMemoryBase + chan * CHANNELBANKSIZE)[pointer + 3];
					pointer += 4;
					v = ((long long)v3<<48) + ((long long)v2<<32) + ((long long)v1<<16) + v0;
					if(addup == 1)
					{
						sum += v;
						histogram[chan][i] = (double) sum;
					}else
     					histogram[chan][i] = (double) v;
     			}
				if(addup == 0)
				{
     				this->scalar->reg.csr |= CSR_ETM;																//fetch totalised number
     				v0 = (this->convMemoryBase + chan * 4)[0];
					v1 = (this->convMemoryBase + chan * 4)[1];
					v2 = (this->convMemoryBase + chan * 4)[2];
					v3 = (this->convMemoryBase + chan * 4)[3];
					v = ((long long)v3<<48) + ((long long)v2<<32) + ((long long)v1<<16) + v0;
					histogram[chan][i] = (double) v;
				}else
					histogram[chan][i] = (double) sum;

				if ((i + 1) < CHANNELBANKSIZE / 4) histogram[chan][i + 1] = 0;
     		}
     	}

        /****************************************/
        /****************************************/
        /********* waveform collection **********/
        /****************************************/
        /****************************************/

        for(chan=0; chan<MAX_CHANNELS; chan++)
        {
        	if(mode == hy8522straightscaler)																	//straight scaler
        	{
        		doCallbacksInt32Array((epicsInt32 *) &basic[chan],
					noofpoints, hy8522Data_, chan);
        		this->isdone_ = true;																			//tell straight scaler read process all finished
        	}
        	else if((mode == hy8522histogramnormal) || (mode == hy8522histogramcoincidence))					//historgram
        	{
        		doCallbacksFloat64Array((epicsFloat64 *) &histogram[chan],
 					noofpoints+1, hy8522Data_, chan);
        	}
		}

        if(mode == hy8522presetscaler) 																			//preset scaler
        {
        	setIntegerParam(scalerDone_, 1);
            this->scalar->reg.overflow = 0xFFFF;																//clear overflow
        }
        else
        {
         	getIntegerParam(hy8522Restart_, &restart);															//fetch restart mode
         	if(restart == hy8522carryonrestart)																					//carry on restart
         	{
            	this->scalar->reg.cycles = 0;																	//clear cycles    
            	getIntegerParam(hy8522NoCycles_, &cycles);
            	this->scalar->reg.cycles = cycles;																//re-write cycles         		
         	}else if(restart == hy8522cleanrestartatINT)
         		resetScaler();																					//reset scaler immediately
        }
        
		setIntegerParam(hy8522TriggerStatus_, FINISHED_AVAILABLE);												//collecting finished. Set trigger status to FINISHED
        this->scalar->reg.csr |= CSR_IE;																		//re-enable interrupt
        ipmIrqCmd(this->carrierNum_, this->ipSlotNum_, 0, ipac_irqEnable);
        callParamCallbacks();
//printf("end csr=%x overflow=%x\n", this->scalar->reg.csr, this->scalar->reg.overflow);
    }
}

