/*******************/
/* System includes */
/*******************/

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <limits.h>

/******************/
/* EPICS includes */
/******************/

/* EPICS includes */
#include <drvIpac.h>
#include <iocsh.h>
#include <epicsExport.h>
#include <epicsThread.h>
#include <epicsString.h>
#include <epicsTimer.h>
#include <epicsExit.h>
#include <epicsMutex.h>
#include <epicsMessageQueue.h>
#include <cantProceed.h>
#include <devLib.h>
#include <errlog.h>
#include <asynPortDriver.h>
#include "scalerRecord.h"
#include <callback.h>

#ifndef DRVHY8522IPASYN_H
#define DRVHY8522IPASYN_H
#endif

/**************************************************************/
/* Few key parameters defining the scaler memory layout. */
#define CHANNELBANKSIZE  		0x10000        	/* maximum channel size: 64K words */
#define MAX_MESSAGES  			100           	/* maximum number of messages */
#define MAX_CHANNELS    		16      		/* number of channels on 8522 card */
#define MAX_COUNT    			0xFFFFFFFF 		/* full 32 bit count */
#define MAX_CARDS          256          /* maxiumum number of cards in a system */

/* Identifiers in PROM. */
#define VITA4       			"VITA 4"
#define HYTECID_LO  			0x0300
#define HYTECID_HI  			0x0080
#define PROM_MODEL  			0x8522

/* define bits in the 8522 CSR */
#define CSR_IRQ					0x0001          /* STAUTS read only: interrupt generated */
#define CSR_R	 				0x0002          /* Reset: write 1 to reset all counters and memeory to 0 */
#define CSR_IE	 				0x0004          /* Enable interrupt when set 1 */
#define CSR_ET 					0x0008          /* Enable hardware trigger from transition board to initiate the cycle */
#define CSR_ER		 			0x0010          /* Enable external hardware reset */
#define CSR_ST	 				0x0020          /* Software trigger initiates a single cycle */
#define CSR_EC 					0x0040          /* Enable coincidence */

#define CSR_TM     				0x0100          /* Set time interval. 1: 1ms, 0: 100us */
#define CSR_MODE_HISTOGRAM		0x0000          /* Mode. 00: histogram mode */
#define CSR_MODE_STRAIGHT		0x0200          /* Mode. 01: straight scaler mode */
#define CSR_MODE_PRESET			0x0400          /* Mode. 10: pre-set scaler mode */
#define CSR_MODE_UNUSED			0x0600          /* not used for the current version */
#define CSR_PR					0x0800			/* Overflow bit when preset value has reaches */
#define CSR_EBA					0x1000			/* Enable external time bin advancing */
#define CSR_TS_25				0x0000			/* 25MHz */
#define CSR_TS_50				0x2000			/* 50MHz */
#define CSR_TS_100				0x4000			/* 100MHz */
#define CSR_TS_200				0x6000			/* 200MHz */
#define CSR_ETM     			0x8000          /* Toggle totalisers and memory. */

#define DIS_INT     			0x0001          /* Disable interrupt. */
#define MODE_HISTOGRAM 			0x0000          /* histogram and straight scaler mode. */
#define MODE_PRESET    			0x0100          /* preset scaler mode */
#define MODE_SET    			0x0080          /* set mode */

/* driver support flags */
/* Bit0 - continuos mode support */
/* Bit1 - trigger mode support */
#define DRIVER_SUPPORT     		0x0007   		/* support histogram mode, coincidence and basic counter mode and  */

#define WAITING_TRIGGER         0           	/* waiting for trigger */
#define TRIGGERRED_COLLECTING   1           	/* triggerred and collecting data */
#define FINISHED_AVAILABLE      2           	/* finished and data available */

#ifndef USHORT 
#define USHORT unsigned short
#endif
#ifndef ULONG
#define ULONG unsigned int
#endif
#ifndef ULONGLONG
#define ULONGLONG unsigned long long
#endif

/*****************************************************************************/
/*                                                                           */
/*                              Data Structures                              */
/*                                                                           */
/*****************************************************************************/


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* ADC Register and PROM Layout                                              */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* The I/O memory is divided into two areas: register and PROM. We separate
 * them out for coding convenience. 
 */

/* This data structure maps the I/O registers */
typedef struct
{									//for histogram						//for preset scaler 
	USHORT reg[MAX_CHANNELS*2];		/* 0 - 31 */						//16 32bit preset value
    USHORT coin[MAX_CHANNELS];     	/* 32 - 47 */						// *** not used ***
    USHORT csr;                		/* 48 csr */						//csr
    USHORT gateintv;          		/* 49 gate interval */				// *** not used ***
    USHORT cycles;               	/* 50 number of cycles */			// *** not used ***
    USHORT gates;                  	/* 51 number of gates */			// *** not used ***
    USHORT overflow;           		/* 52  */							//overflow
    USHORT triggerno;              	/* 53 triggers count */				//preset enable mask register
    USHORT vector	;             	/* 54 vector */						//vector
    USHORT intdisable;  			/* 55 */							//interrupt enable/disable
    USHORT pointerl;  				/* 56 Memory pointer low */			// *** not used ***
    USHORT pointerh;           		/* 57 Memory pointer high */		// *** not used ***
    USHORT disablechl;           	/* 58 MASK */						//ARM channel
    USHORT function;           		/* 59 function register */			//change of function register
    USHORT presetl;           		/* 60  *** not used *** */			// *** not used ***
    USHORT preseth;           		/* 61  *** not used *** */			// *** not used ***
    USHORT intextclock;           	/* 62 internal/external clock */	//internal/external source
} REGISTERS;


/* This data structure maps the image of the ID PROM.  The PROM
 * starts at offset 0x80 from the base of the register memory. */
typedef struct
{
    char vita4[6];                  /* 0x80 */
    USHORT hytecIdHigh;             /* 0x86 */
    USHORT hytecIdLow;              /* 0x88 */
    USHORT model;                   /* 0x8A */
    USHORT revision;                /* 0x8C */
    USHORT unused1;                 /* 0x8E */
    USHORT driverIdHigh;        	/* 0x90 */
    USHORT driverIdLow;             /* 0x92 */
    USHORT flags;                   /* 0x94 */
    USHORT byteCount;               /* 0x96 */
    USHORT unused2;                 /* 0x98 */
    USHORT serialNumber;            /* 0x9A */
} ID_PROM;

/* Now put them together, the I/O memory map. */
typedef struct    	
{
    REGISTERS reg;                              /* 0x00 - 0x2F */
    char unused[0x80-sizeof(REGISTERS)];        /* 0x30 - 0x80 */
    ID_PROM prom;                               /* 0x80 - 0x9F */
} IO_MEMORY;

/* enumerated type of the driver's operational modes */
typedef enum{		
	hy8522histogramnormal,
	hy8522histogramcoincidence,
	hy8522straightscaler,
	hy8522presetscaler
} hy8522Modes;

/* enumerated type of the driver's operational modes */
typedef enum{		
	hy8522singleshot,
	hy8522carryonrestart,
	hy8522cleanrestart,
	hy8522cleanrestartatINT
} restartMode;


typedef struct {
    scalerRecord *psr;
    asynUser *pasynUser[MAX_CHANNELS];
    asynInt32 *pasynInt32;
    void *asynInt32Pvt;
    asynInt32Array *pasynInt32Array;
    void *asynInt32ArrayPvt;
    asynDrvUser *pasynDrvUser;
    void  *asynDrvUserPvt;
    int channelsCommand;
    int resetCommand;
    int readCommand;
    int presetCommand;
    int armCommand;
    int doneCommand;
    int done;
    void *registrarPvt;
    CALLBACK *pcallback;
} scalerAsynPvt;

/*****************************************************************************/
/*                                                                           */
/*                      Interrupt/Queue Routines	                         */
/*                                                                           */
/*****************************************************************************/

extern "C" {
static void rebootCallbackC     ( void *drvPvt );
static void INTSRC              ( int drvPvt ); 	/* Interrupt Service Routine */
static void intQueuedTaskC      ( void *drvPvt );
}

/** This is the class definition for the Hy8522Asyn class
  */
class Hy8522Asyn : public asynPortDriver {
public:
	Hy8522Asyn(char *portName, int carrierNum, int ipSlotNum, int vectorNum, int freq, int sourceMask, int hwtrigger, int hwreset);

    /* These are the methods that we override from asynPortDriver */
    virtual asynStatus writeInt32(asynUser *pasynUser, epicsInt32 value);
    virtual asynStatus readInt32(asynUser *pasynUser, epicsInt32 *value);
    virtual asynStatus readInt32Array(asynUser *pasynUser, epicsInt32 *value,
                                        size_t nElements, size_t *nIn);    
    virtual asynStatus readFloat64Array(asynUser *pasynUser, epicsFloat64 *value,
            size_t nElements, size_t *nIn);
    virtual void report(FILE *fp, int details);
    
    int Hy8522HistogramInit(int gateIntv, int timeunit, int nocycles, int nogates, int mask, int extadvance, int accumulate);	
    int Hy8522CoinInit(int gateIntv, int timeunit, int nocycles, int nogates, int mask, int extadvance, int accumulate);	
    int Hy8522StraightInit(int gateIntv, int timeunit, int nocycles, int mask);	
    int Hy8522PresetInit(int mask);	
    int Hy8522InitCoin0_7(int coin0, int coin1, int coin2, int coin3, 
    		int coin4, int coin5, int coin6, int coin7);	
    int Hy8522InitCoin8_15(int coin8, int coin9, int coin10, int coin11, 
    		int coin12, int coin13, int coin14, int coin15);	

    void rebootCallback();
	void intSR(); 			/* Interrupt Service Routine */
	void intQueuedTask();

protected:
    int hy8522Data_;
    #define FIRST_8522_PARAM hy8522Data_
	int hy8522Coin_;
	int hy8522Mode_;
	int hy8522SoftTrigger_;
	int hy8522HWTrigger_;
	int hy8522HWReset_;
	int hy8522GateIntv_;
	int hy8522NoCycles_;
	int hy8522NoGates_;
	int hy8522DiableInput_;
	int hy8522TimeUnit_;
	int hy8522ExtAdvance_;
	int hy8522Addup_;
	int hy8522Frequency_;
	int hy8522Source_;
	int hy8522Restart_;

	int scalerReset_;
	int scalerChannels_;
	int scalerRead_;
	int scalerReadSingle_;
	int scalerPresets_;
	int scalerArm_;
	int scalerDone_;

	int hy8522MemPointer_;
	int hy8522NoTriggers_;
	int hy8522TriggerStatus_;
	int hy8522FWVersion_;
	int hy8522DriverVersion_;
	int hy8522Support_;
    #define LAST_8522_PARAM hy8522Support_

private:
	void setupInterrupt            	();
	void setupScaler            	();
	asynStatus setMode          	( int mode );
	asynStatus setTimeUint      	( int unit );
	asynStatus setEnableHWTrigger   ( int HWtrigger );
	asynStatus setEnableHWReset		( int HWreset );
	asynStatus setMask         		( int mask );
	asynStatus setCoin        		( int channel, int coin );
	asynStatus setGateInterval		( int gateintv );
	asynStatus setNoOfCycles		( int nocycles );
	asynStatus setNoOfGates			( int nogates );
	asynStatus softwareTrigger		();
	asynStatus setExtAdvance		( int extadv );
	asynStatus setAddupMode			( int addup );
	asynStatus setFrequency			( int freq );
	asynStatus setSource			( int source );
	asynStatus restart				( int mode );
	asynStatus readScalers			();
	asynStatus resetScaler			();
	asynStatus setScalerPresets		();
	asynStatus startScalers			();
	asynStatus stopScalers			();

	volatile IO_MEMORY 	*scalar;			/* pointer to IP card A16 space registers, ID and calibration data */ 
	volatile USHORT 	*convMemoryBase;   	/* pointer to IP card memory space */

    ULONG 	basic[MAX_CHANNELS][CHANNELBANKSIZE/2+1];		/* basical scaler data. 32K x 32bit per channel. For I/O Intr Int32Array */
    double 	histogram[MAX_CHANNELS][CHANNELBANKSIZE/4+1];	/* histogram data. 16k x 64bit per channel for I/O Intr Float64Array */

	/* Storage for the current Configuration parameters */
    USHORT	carrierNum_;		            /* carrier index number when adding carrier */
    USHORT	ipSlotNum_;		                /* IP card slot number on IP carrier - boot time configuration only */
    USHORT	vectorNum_;		                /* Interrupt vector number - boot time configuration only */

/* driver specific parameters */
    int		rebooting_;		                /* Flag to indicate is the IOC is currently re-booting */
    bool	isdone_;						/* finish flag for straight scaler mode */
    
    unsigned int presetmask;				/* preset mask */
    int intv_tp;							/* TP interval value */
    int intv_tp1;							/* TP1 interval value */ 
	scalerAsynPvt *pPvt;					/* asynScaler device pointer */
	scalerRecord *psr;						/* scaler record */

/* Work Queue Stuff */
    epicsMessageQueueId intMsgQId;
    int messagesSent;                       /* for report */
    int messagesFailed;                     /* for report */
};

#define hy8522DataString	        "DATA"
#define hy8522CoincidenceString	    "COIN"
#define hy8522ModeString	    	"MODE"
#define hy8522HWTriggerString	    "ENABLEHWTRIGGER"
#define hy8522HWResetString	    	"ENABLEHWRESET"
#define hy8522GateIntervalString	"GATEINTERVAL"
#define hy8522NoCyclesString		"NOCYCLES"
#define hy8522NoGatesString		    "NOGATES"
#define hy8522SoftTriggerString		"SOFTTRIGGER"
#define hy8522DisableInputString	"INPUTMASK"
#define hy8522TimeUnitString		"TIMEUNIT"
#define hy8522ExtAdvanceString		"EXTADVANCE"
#define hy8522AddupString			"ADDUPHISTOGRAM"
#define hy8522FrequencyString		"FREQUENCY"
#define hy8522SourceString			"SOURCE"
#define hy8522RestartString			"RESTART"

#define hy8522MemPointerString		"MEMPOINTER"
#define hy8522NoTriggersString		"NOTRIGGERS"
#define hy8522TriggerStatusString	"TRIGGERSTATE"
#define hy8522FWVersionString		"FWVERSION"
#define hy8522DriverVersionString	"DRIVERVERSION"
#define hy8522SupportString			"SUPPORT"

#define NUM_PARAMS (&LAST_8522_PARAM - &FIRST_8522_PARAM + 1)

/** Configuration command, called directly or from iocsh */
/* 
 * Where: -
 *     portName = asyn port name
 *   carrierNum = carrier index number when adding a carrier card by ipacAddHy8002 call
 *    ipSlotNum = IP slot number on the carrier board (0 = A, 1 = B, 2 = C, 3 = D, 4 = E, 5 = F),
 *    vectorNum = interrupt vector (0 = find one for me),
 *	 	   freq = For internal clock. 0: 25MHz, 1:50MHz, 2:100MHz, 3:200MHz
 *	 sourceMask = bit mask for channel clock source. 0: external. 1: internal defined by freq param
 *	 hwtrigger  = 1 = enable hardware trigger, 0 = only software trigger
 *    hwreset   = 1 = allow hardware reset, 0 = only software reset
 */
extern "C" int Hy8522AsynInit(char *portName, int vmeSlotNum, int ipSlotNum, int vectorNum, 
				int freq, int sourceMask, int hwtrigger, int hwreset)
{
    new Hy8522Asyn(portName, vmeSlotNum, ipSlotNum, vectorNum, 
    		freq, sourceMask, hwtrigger, hwreset);
    return(asynSuccess);
}

/** Normal histogram configuration command, called directly or from iocsh */
/* 
 * Where: -
 *     portName = asyn port name
 *     gateIntv = When extadvance setting is 0 , this is the gate interval as below: 
 *     					If timeunit = 0, i.e. 1ms, this defines the interval in mini-second 
 *     			  		If timeunit = 1, i.e. 100us, this defines the interval in 100us steps. 
 *     			  When extadvance = 1, it is number of pulses to advance the memory location
 *     timeunit = Define interval time unit. 0 = 1ms, 1 = 100us
 *     nocycles	= Define the number of cycles. If it is set to 0, the scaler doesn't do anything.
 *      nogates = Define the number of gates (bins), i.e. How many gates in a cycle. Each gate 
 *      				stores counting of pulses in one memory location. Next cycle adds up to 
 *      				the same locatoin to form histogram.
 *         mask = bit mask to enable/disable channels. 0 = enable, 1 = disable.
 *   extadvance = Memory advance control. 0 = advanced by interval timer, 1 = advanced by external hardware pulses 
 *   accumulate = Define the returning waveform of the histogram. =0 return normal histogram. = 1 return accumulated 
 *				  		waveform -- each bin also accumulates all previous bins.
 */
extern "C" int Hy8522HistogramMode(char *portName, int gateIntv, int timeunit, int nocycles, int nogates, 
		int mask, int extadvance, int accumulate)
{
    asynStatus status;
    asynInterface *pasynInterface;
    asynUser *pasynUser;
    Hy8522Asyn *pHy8522 = NULL;

    pasynUser = pasynManager->createAsynUser(0, 0);
    status = pasynManager->connectDevice(pasynUser, portName, 0);
    if (status != asynSuccess) {
        errlogPrintf("Hy8522AsynExtInit, error in connectDevice %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pasynInterface = pasynManager->findInterface(pasynUser, asynInt32Type, 1);
    if (!pasynInterface) {
        errlogPrintf("Hy8522AsynExtInit cannot find Float64 interface %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pHy8522 = (Hy8522Asyn*) pasynInterface->drvPvt;
    pHy8522->Hy8522HistogramInit(gateIntv, timeunit, nocycles, nogates, mask, extadvance, accumulate);
    return(asynSuccess);
} 

/** Coincidence histogram configuration command, called directly or from iocsh */
/* 
 * Where: -
 *     portName = asyn port name
 *     gateIntv = When extadvance setting is 0 , this is the gate interval as below: 
 *     					If timeunit = 0, i.e. 1ms, this defines the interval in mini-second 
 *     			  		If timeunit = 1, i.e. 100us, this defines the interval in 100us steps. 
 *     			  When extadvance = 1, it is number of pulses to advance the memory location
 *     timeunit = Define interval time unit. 0 = 1ms, 1 = 100us
 *     nocycles	= Define the number of cycles. If it is set to 0, the scaler doesn't do anything.
 *      nogates = Define the number of gates (bins), i.e. How many gates in a cycle. Each gate 
 *      				stores counting of pulses in one memory location. Next cycle adds up to 
 *      				the same locatoin to form histogram.
 *         mask = bit mask to enable/disable channels. 0 = enable, 1 = disable.
 *   extadvance = Memory advance control. 0 = advanced by interval timer, 1 = advanced by external hardware pulses 
 *   accumulate = Define the returning waveform of the histogram. =0 return normal histogram. = 1 return accumulated 
 *				  		waveform -- each bin also accumulates all previous bins.
 */
extern "C" int Hy8522CoincidenceMode(char *portName, int gateIntv, int timeunit, int nocycles, int nogates, 
		int mask, int extadvance, int accumulate)
{
    asynStatus status;
    asynInterface *pasynInterface;
    asynUser *pasynUser;
    Hy8522Asyn *pHy8522 = NULL;

    pasynUser = pasynManager->createAsynUser(0, 0);
    status = pasynManager->connectDevice(pasynUser, portName, 0);
    if (status != asynSuccess) {
        errlogPrintf("Hy8522AsynExtInit, error in connectDevice %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pasynInterface = pasynManager->findInterface(pasynUser, asynInt32Type, 1);
    if (!pasynInterface) {
        errlogPrintf("Hy8522AsynExtInit cannot find Float64 interface %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pHy8522 = (Hy8522Asyn*) pasynInterface->drvPvt;
    pHy8522->Hy8522CoinInit(gateIntv, timeunit, nocycles, nogates, mask, extadvance, accumulate);
    return(asynSuccess);
} 

/** Straight scaler configuration command, called directly or from iocsh */
/* 
 * Where: -
 *     portName = asyn port name
 *     gateIntv = Define gate interval as below: 
 *     					If timeunit = 0, i.e. 1ms, this defines the interval in mini-second 
 *     			  		If timeunit = 1, i.e. 100us, this defines the interval in 100us steps. 
 *     timeunit = Define interval time unit. 0 = 1ms, 1 = 100us
 *     nocycles	= Define the number of trigger cycles. Up to 32K.
 *         mask = bit mask to enable/disable channels. 0 = enable, 1 = disable.
 */
extern "C" int Hy8522StraightMode(char *portName, int gateIntv, int timeunit, int nocycles, int mask)
{
    asynStatus status;
    asynInterface *pasynInterface;
    asynUser *pasynUser;
    Hy8522Asyn *pHy8522 = NULL;

    pasynUser = pasynManager->createAsynUser(0, 0);
    status = pasynManager->connectDevice(pasynUser, portName, 0);
    if (status != asynSuccess) {
        errlogPrintf("Hy8522AsynExtInit, error in connectDevice %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pasynInterface = pasynManager->findInterface(pasynUser, asynInt32Type, 1);
    if (!pasynInterface) {
        errlogPrintf("Hy8522AsynExtInit cannot find Float64 interface %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pHy8522 = (Hy8522Asyn*) pasynInterface->drvPvt;
    pHy8522->Hy8522StraightInit(gateIntv, timeunit, nocycles, mask);
    return(asynSuccess);
} 

/** Preset scaler configuration command, called directly or from iocsh */
/* 
 * Where: -
 *     portName = asyn port name
 *      armMask = Define ARM bit. Set to 1 to ARM the correspondent channel
 */
extern "C" int Hy8522PresetMode(char *portName,int armMask)
{
    asynStatus status;
    asynInterface *pasynInterface;
    asynUser *pasynUser;
    Hy8522Asyn *pHy8522 = NULL;

    pasynUser = pasynManager->createAsynUser(0, 0);
    status = pasynManager->connectDevice(pasynUser, portName, 0);
    if (status != asynSuccess) {
        errlogPrintf("Hy8522AsynExtInit, error in connectDevice %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pasynInterface = pasynManager->findInterface(pasynUser, asynInt32Type, 1);
    if (!pasynInterface) {
        errlogPrintf("Hy8522AsynExtInit cannot find Float64 interface %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pHy8522 = (Hy8522Asyn*) pasynInterface->drvPvt;
    pHy8522->Hy8522PresetInit(armMask);
    return(asynSuccess);
} 

extern "C" int Hy8522AsynInitCoin0_7(char *portName, int coin0, int coin1, int coin2, 
		int coin3, int coin4, int coin5, int coin6, int coin7)
{
    asynStatus status;
    asynInterface *pasynInterface;
    asynUser *pasynUser;
    Hy8522Asyn *pHy8522 = NULL;

    pasynUser = pasynManager->createAsynUser(0, 0);
    status = pasynManager->connectDevice(pasynUser, portName, 0);
    if (status != asynSuccess) {
        errlogPrintf("Hy8522AsynExtInit, error in connectDevice %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pasynInterface = pasynManager->findInterface(pasynUser, asynInt32Type, 1);
    if (!pasynInterface) {
        errlogPrintf("Hy8522AsynExtInit cannot find Float64 interface %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pHy8522 = (Hy8522Asyn*) pasynInterface->drvPvt;
    pHy8522->Hy8522InitCoin0_7(coin0, coin1, coin2, coin3, coin4, coin5, coin6, coin7);
    return(asynSuccess);
} 

extern "C" int Hy8522AsynInitCoin8_15(char *portName, int coin8, int coin9, int coin10, 
		int coin11, int coin12, int coin13, int coin14, int coin15)
{
    asynStatus status;
    asynInterface *pasynInterface;
    asynUser *pasynUser;
    Hy8522Asyn *pHy8522 = NULL;

    pasynUser = pasynManager->createAsynUser(0, 0);
    status = pasynManager->connectDevice(pasynUser, portName, 0);
    if (status != asynSuccess) {
        errlogPrintf("Hy8522AsynExtInit, error in connectDevice %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pasynInterface = pasynManager->findInterface(pasynUser, asynInt32Type, 1);
    if (!pasynInterface) {
        errlogPrintf("Hy8522AsynExtInit cannot find Float64 interface %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pHy8522 = (Hy8522Asyn*) pasynInterface->drvPvt;
    pHy8522->Hy8522InitCoin8_15(coin8, coin9, coin10, coin11, coin12, coin13, coin14, coin15);
    return(asynSuccess);
} 

// General Asyn Init
static const iocshArg initArg0 = { "Port name",iocshArgString};
static const iocshArg initArg1 = { "vmeSlotNum",iocshArgInt};
static const iocshArg initArg2 = { "ipSlotNum",iocshArgInt};
static const iocshArg initArg3 = { "vectorNum",iocshArgInt};
static const iocshArg initArg4 = { "frequency", iocshArgInt};
static const iocshArg initArg5 = { "source mask", iocshArgInt};
static const iocshArg initArg6 = { "hardware trigger", iocshArgInt};
static const iocshArg initArg7 = { "hardware reset", iocshArgInt};
static const iocshArg * const initArgs[] = {&initArg0,
                                             &initArg1,
                                             &initArg2,
                                             &initArg3,
                                             &initArg4,
                                             &initArg5,
                                             &initArg6,
                                             &initArg7};
static const iocshFuncDef initFuncDef = {"Hy8522AsynInit",8,initArgs};
static void initCallFunc(const iocshArgBuf *args)
{
    Hy8522AsynInit(args[0].sval, args[1].ival, args[2].ival, args[3].ival,
					args[4].ival, args[5].ival, args[6].ival, args[7].ival);
}

// Normal Histogram Mode init
static const iocshArg initArg0His = { "Port name",iocshArgString};
static const iocshArg initArg1His = { "gateIntv",iocshArgInt};
static const iocshArg initArg2His = { "timeunit", iocshArgInt};
static const iocshArg initArg3His = { "nocycles",iocshArgInt};
static const iocshArg initArg4His = { "nogates",iocshArgInt};
static const iocshArg initArg5His = { "mask",iocshArgInt};
static const iocshArg initArg6His = { "extadvance",iocshArgInt};
static const iocshArg initArg7His = { "addup",iocshArgInt};
static const iocshArg * const initArgsHis[] = {&initArg0His,
                                             &initArg1His,
                                             &initArg2His,
                                             &initArg3His,
                                             &initArg4His,
                                             &initArg5His,
                                             &initArg6His,
                                             &initArg7His};
static const iocshFuncDef initFuncDefHis = {"Hy8522HistogramMode",8,initArgsHis};
static void initCallFuncHis(const iocshArgBuf *args)
{
	Hy8522HistogramMode(args[0].sval, args[1].ival, args[2].ival, args[3].ival,
					args[4].ival, args[5].ival, args[6].ival, args[7].ival);
}

// Coincidence Histogram Mode init
static const iocshArg initArg0Coin = { "Port name",iocshArgString};
static const iocshArg initArg1Coin = { "gateIntv",iocshArgInt};
static const iocshArg initArg2Coin = { "timeunit", iocshArgInt};
static const iocshArg initArg3Coin = { "nocycles",iocshArgInt};
static const iocshArg initArg4Coin = { "nogates",iocshArgInt};
static const iocshArg initArg5Coin = { "mask",iocshArgInt};
static const iocshArg initArg6Coin = { "extadvance",iocshArgInt};
static const iocshArg initArg7Coin = { "addup",iocshArgInt};
static const iocshArg * const initArgsCoin[] = {&initArg0Coin,
                                             &initArg1Coin,
                                             &initArg2Coin,
                                             &initArg3Coin,
                                             &initArg4Coin,
                                             &initArg5Coin,
                                             &initArg6Coin,
                                             &initArg7Coin};
static const iocshFuncDef initFuncDefCoin = {"Hy8522CoincidenceMode",8,initArgsCoin};
static void initCallFuncCoin(const iocshArgBuf *args)
{
	Hy8522CoincidenceMode(args[0].sval, args[1].ival, args[2].ival, args[3].ival,
					args[4].ival, args[5].ival, args[6].ival, args[7].ival);
}

// Extension 1 init for coincidence 0 ~ 7
static const iocshArg initArg0Ext07 = { "Port name",iocshArgString};
static const iocshArg initArg1Ext07 = { "coin0",iocshArgInt};
static const iocshArg initArg2Ext07 = { "coin1",iocshArgInt};
static const iocshArg initArg3Ext07 = { "coin2",iocshArgInt};
static const iocshArg initArg4Ext07 = { "coin3",iocshArgInt};
static const iocshArg initArg5Ext07 = { "coin4", iocshArgInt};
static const iocshArg initArg6Ext07 = { "coin5", iocshArgInt};
static const iocshArg initArg7Ext07 = { "coin6", iocshArgInt};
static const iocshArg initArg8Ext07 = { "coin7",iocshArgInt};
static const iocshArg * const initArgsExtCoin1[] = {&initArg0Ext07,
                                             &initArg1Ext07,
                                             &initArg2Ext07,
                                             &initArg3Ext07,
                                             &initArg4Ext07,
                                             &initArg5Ext07,
                                             &initArg6Ext07,
                                             &initArg7Ext07,
                                             &initArg8Ext07};
static const iocshFuncDef initFuncDefCoin1 = {"Hy8522AsynInitCoin1",9,initArgsExtCoin1};
static void initCallFuncCoin1(const iocshArgBuf *args)
{
    Hy8522AsynInitCoin0_7(args[0].sval, args[1].ival, args[2].ival, args[3].ival,
					args[4].ival, args[5].ival, args[6].ival, args[7].ival,
					args[8].ival);
}

// Extension 2 init for coincidence 8 ~ 15
static const iocshArg initArg0Ext815 = { "Port name",iocshArgString};
static const iocshArg initArg1Ext815 = { "coin8",iocshArgInt};
static const iocshArg initArg2Ext815 = { "coin9",iocshArgInt};
static const iocshArg initArg3Ext815 = { "coin10",iocshArgInt};
static const iocshArg initArg4Ext815 = { "coin11",iocshArgInt};
static const iocshArg initArg5Ext815 = { "coin12", iocshArgInt};
static const iocshArg initArg6Ext815 = { "coin13", iocshArgInt};
static const iocshArg initArg7Ext815 = { "coin14", iocshArgInt};
static const iocshArg initArg8Ext815 = { "coin15",iocshArgInt};
static const iocshArg * const initArgsExtCoin2[] = {&initArg0Ext815,
                                             &initArg1Ext815,
                                             &initArg2Ext815,
                                             &initArg3Ext815,
                                             &initArg4Ext815,
                                             &initArg5Ext815,
                                             &initArg6Ext815,
                                             &initArg7Ext815,
                                             &initArg8Ext815};
static const iocshFuncDef initFuncDefCoin2 = {"Hy8522AsynInitCoin2",9,initArgsExtCoin2};
static void initCallFuncCoin2(const iocshArgBuf *args)
{
    Hy8522AsynInitCoin8_15(args[0].sval, args[1].ival, args[2].ival, args[3].ival,
					args[4].ival, args[5].ival, args[6].ival, args[7].ival,
					args[8].ival);
}

// Straight Scaler Mode init
static const iocshArg initArg0Straight = { "Port name",iocshArgString};
static const iocshArg initArg1Straight = { "gateIntv",iocshArgInt};
static const iocshArg initArg2Straight = { "timeunit", iocshArgInt};
static const iocshArg initArg3Straight = { "nocycles",iocshArgInt};
static const iocshArg initArg4Straight = { "mask",iocshArgInt};
static const iocshArg * const initArgsStraight[] = {&initArg0Straight,
                                             &initArg1Straight,
                                             &initArg2Straight,
                                             &initArg3Straight,
                                             &initArg4Straight};
static const iocshFuncDef initFuncDefStraight = {"Hy8522StraightMode",5,initArgsStraight};
static void initCallFuncStraight(const iocshArgBuf *args)
{
	Hy8522StraightMode(args[0].sval, args[1].ival, args[2].ival, args[3].ival,
					args[4].ival);
}

// Preset Scaler Mode init
static const iocshArg initArg0Preset = { "Port name",iocshArgString};
static const iocshArg initArg1Preset = { "armMask",iocshArgInt};
static const iocshArg * const initArgsPreset[] = {&initArg0Preset,
                                             &initArg1Preset};
static const iocshFuncDef initFuncDefPreset = {"Hy8522PresetMode",2,initArgsPreset};
static void initCallFuncPreset(const iocshArgBuf *args)
{
	Hy8522PresetMode(args[0].sval, args[1].ival);
}


void drvHy8522AsynRegister(void)
{
    iocshRegister(&initFuncDef,initCallFunc);
    iocshRegister(&initFuncDefHis,initCallFuncHis);
    iocshRegister(&initFuncDefCoin,initCallFuncCoin);
    iocshRegister(&initFuncDefCoin1,initCallFuncCoin1);
    iocshRegister(&initFuncDefCoin2,initCallFuncCoin2);
    iocshRegister(&initFuncDefStraight,initCallFuncStraight);
    iocshRegister(&initFuncDefPreset,initCallFuncPreset);
}

extern "C" {
epicsExportRegistrar(drvHy8522AsynRegister);
}

